import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Linear.LLQueue import LLQueue
from datastructures.Nodes.SinglyNode import Node

testNode1 = Node(6)
testNode2 = Node(1)
testNode3 = Node(2)
testNode4 = Node(7)
testNode5 = Node(10)   
testNode6 = Node(3)
testNode6_2 = Node(3)
testNode7 = Node(4)

@pytest.mark.test_constructor
def test_constructor():
    testQueue = LLQueue()
    assert testQueue.head == None
    assert testQueue.tail == None

    testQueue2 = LLQueue(Node(2))
    assert testQueue2.head.data == 2
    assert testQueue2.tail.data == 2

@pytest.mark.test_enqueue
def test_enqueue():
    testQueue = LLQueue()
    testQueue.enqueue(testNode1)
    assert testQueue.head.data == 6
    testQueue.enqueue(testNode2)
    testQueue.enqueue(testNode3) 
    testQueue.enqueue(testNode4)
    testQueue.enqueue(testNode5)
    assert testQueue.head.next.next.data == 2
    testQueue.enqueue(testNode6)
    testQueue.enqueue(testNode7) # test enqueue at tail
    currentTestNode = testQueue.head 
    testList = []
    for i in range(0, testQueue.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [6, 1, 2, 7, 10, 3, 4]
    testQueue.enqueue(testNode6_2)
    testList = []
    currentTestNode = testQueue.head
    for i in range(0, testQueue.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList == [6, 1, 2, 7, 10, 3, 4, 3]

def setup():
    testQueue = LLQueue()
    testQueue.enqueue(testNode1) # 6
    testQueue.enqueue(testNode2) # 1
    testQueue.enqueue(testNode3) # 2
    testQueue.enqueue(testNode4) # 7
    testQueue.enqueue(testNode5) # 10
    testQueue.enqueue(testNode6) # 3
    testQueue.enqueue(testNode6_2) # 3
    testQueue.enqueue(testNode7) # 4
    return testQueue

@pytest.mark.test_dequeue
def test_dequeue():
    testQueue = setup()
    testQueue.dequeue()
    assert testQueue.head.data != 6
    testQueue.dequeue()
    testQueue.dequeue()
    currentTestNode = testQueue.head
    testList = []
    for i in range(0, testQueue.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [7, 10, 3, 3, 4]
    testQueue.clear()
    assert testQueue.size == 0 and testQueue.head == None

@pytest.mark.test_search
def test_Search():
    testQueue = setup()
    testNode6Search = testQueue.search(testNode7)
    assert testNode6Search == testNode7
    nonListNode = Node(99)
    testNonListNode = testQueue.search(nonListNode)
    assert testNonListNode == None

@pytest.mark.test_print
def test_print():
    testQueue = setup()
    captured_output = io.StringIO()
    sys.stdout = captured_output
    testQueue.print()
    actual_output = captured_output.getvalue()

    sys.stdout = sys.__stdout__

    expected_output = "The size of this singly linked list queue is 8.\nThis singly linked list queue is not sorted.\nThis singly linked list queue contains:\n6\n1\n2\n7\n10\n3\n3\n4\n"
    assert actual_output ==  expected_output

@pytest.mark.test_detreimentFunctionality
def test_detrimentFunctionality():
    testQueue = LLQueue(testNode7)
    testQueue.insertHead(testNode1)
    testQueue.insert(testNode2,1)
    testQueue.insertTail(testNode3)
    testQueue.deleteHead()
    testQueue.deleteTail()
    testQueue.delete(testNode7)
    testQueue.insertTail(Node(2))
    testQueue.enqueue(Node(2))
    testQueue.sort()
    testQueue.sortedInsert(Node(3))
    testList = []
    currentTestNode = testQueue.head
    for i in range(testQueue.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4, 2]


import sys
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Linear.SinglyLL import singlyLL
from Nodes.SinglyNode import Node

class LLStack(singlyLL):
    def __init__(self, newNode = None):
        super().__init__(newNode)
    
    def push(self, newNode): # Renamed push
        super().insertHead(newNode)

    def insertHead(self, newNode): # For overriding superclass method (now renamed to push)
        return

    def insertTail(self, newNode): # Do nothing
        return

    def insert(self, newNode, position): # Do nothing
        return

    def sortedInsert(self, newNode): # Do nothing
        return

    def isSorted(self):
        return super().isSorted()

    def sort(self):
        return

    def search(self, newNode):
        return (super().search(newNode))

    def pop(self): # Renamed pop
        super().deleteHead()
    
    def deleteHead(self): # For overriding superclass method
        return

    def deleteTail(self): # Do nothing
        return

    def delete(self, delNode): # Do nothing
        return

    def clear(self):
        super().clear()

    def print(self):
        print(f'The size of this singly linked list stack is {self.size}.')
        if self.isSorted():
            print("This singly linked list stack is sorted.")
        if not self.isSorted():
            print("This singly linked list stack is not sorted.")
        print("This singly linked list stack contains:")
        current = self.head
        while current != None:
            print(current.data)
            current = current.next

# def main():
#     # test = LLStack(Node(2))
#     # test.push(Node(1))
#     # test.push(Node(3))
#     # test.pop()
#     # test.clear()
#     # newObj = (test.search(Node(1)))
#     # # print(newObj.data)
#     # test.print()

# if __name__ == "__main__":
#     main()






class TNode():
    def __init__(self, data = None, parent = None, leftChild = None, rightChild = None, balance = None):
    # def __init__(self, data = None, balance = None, parent = None, leftChild = None, rightChild = None):
        if parent is None and leftChild is None and rightChild is None and balance == None:
            self.data = data
            self.parent = None
            self.leftChild = None
            self.rightChild = None 
            self.balance = 0
        else:
            self.data = data
            self.parent = parent
            self.leftChild = leftChild
            self.rightChild = rightChild
            self.balance = 0
    
    def setData(self, newData):
        self.data = newData
    
    def getData(self):
        return self.data
    
    def setBalance(self, newBalance):
        self.balance = newBalance
    
    def getBalance(self):
        return self.balance
    
    def setParent(self, newParent):
        self.parent = newParent
    
    def getParent(self):
        return self.parent
    
    def setLeftChild(self, newLeftChild):
        self.leftChild = newLeftChild
    
    def getLeftChild(self):
        return self.leftChild
    
    def setRightChild(self, newRightChild):
        self.rightChild = newRightChild
    
    def getRightChild(self):
        return self.rightChild
    
    def print(self):
        print(f'The balance of this node is {self.balance}.')
        print(f'This node contains value of {self.toString()}.')
        print(f'This node\'s parent is {self.parent}')
        print(f'This node\'s left child is {self.leftChild}')
        print(f'This node\'s right child is {self.rightChild}')
    
    def toString(self):
        return str(self.data)
    

    




import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Linear.DoublyCLL import doublyCLL
from datastructures.Nodes.DoublyNode import Node

testNode1 = Node(6)
testNode2 = Node(1)
testNode3 = Node(2)
testNode4 = Node(7)
testNode5 = Node(10)   
testNode6 = Node(3)
testNode6_2 = Node(3)
testNode7 = Node(4)

@pytest.mark.test_constructor
def test_constructor():
    testDCLL = doublyCLL()
    assert testDCLL.head == None
    assert testDCLL.tail == None
    
    testDCLL2 = doublyCLL(testNode1)
    assert testDCLL2.head.data == 6
    assert testDCLL2.tail.data == 6
    assert testDCLL2.size == 1
    assert testDCLL2.tail.next.data == 6
    assert testDCLL2.head.prev.data == 6

@pytest.mark.test_insert
def test_insert():
    testDCLL = doublyCLL()
    testDCLL.insertHead(testNode1)
    assert testDCLL.head.next.data == 6
    testDCLL.insertHead(testNode2)
    testDCLL.insertTail(testNode3) 
    testDCLL.insertTail(testNode4)
    assert testDCLL.tail.data == 7
    testDCLL.insert(testNode5,3)
    assert testDCLL.head.next.next.data == 10
    assert testDCLL.tail.next.data == 1 # Added this additional test to ensure circular with multiple elements
    testDCLL.insert(testNode6,5)
    testDCLL.insert(Node(11),7)
    assert testDCLL.tail.data == 11
    testDCLL.insert(testNode7,1) # test insert at head
    currentTestNode = testDCLL.head 
    testList = []
    for i in range(0, testDCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4, 1, 6, 10, 2, 3, 7, 11]
    testDCLL.insertTail(testNode6_2)
    currentTestNode = testDCLL.head 
    testList = []
    for i in range(0, testDCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4, 1, 6, 10, 2, 3, 7, 11, 3]
    assert testDCLL.tail.next.data == 4

    testDCLL.sortedInsert(Node(0))
    testDCLL.sortedInsert(Node(12))
    currentTestNode2 = testDCLL.head 
    testList2 = []
    for i in range(0, testDCLL.size):
        testList2.append(currentTestNode2.data)
        currentTestNode2 = currentTestNode2.next
    assert testList2 == [0, 1, 2, 3, 3, 4, 6, 7, 10, 11, 12]
    

    testDCLL2 = doublyCLL()
    testDCLL2.sortedInsert(Node(2))
    assert testDCLL2.head.data == 2
    assert testDCLL2.tail.data == 2
    assert testDCLL2.head.prev.data == 2
    assert testDCLL2.tail.next.data == 2
    
def setup():
    testDCLL = doublyCLL()
    testDCLL.insertTail(testNode1)
    testDCLL.insertTail(testNode2)
    testDCLL.insertTail(testNode3) 
    testDCLL.insertTail(testNode4)
    testDCLL.insertTail(testNode5)
    testDCLL.insertTail(testNode6)
    testDCLL.insertTail(testNode6_2)
    testDCLL.insertTail(testNode7)
    return testDCLL

@pytest.mark.test_isDoubly
def test_isDoubly():
    testDCLL = setup()
    nextNode = testDCLL.head.next
    prevNode = nextNode.prev
    assert prevNode.data == testDCLL.head.data


@pytest.mark.test_isCircular
def test_isCircular():
    testDCLL = setup()
    circTestNode = testDCLL.tail
    circTestNode = circTestNode.next
    assert circTestNode.data == testDCLL.head.data


@pytest.mark.test_delete
def test_delete():
    testDCLL = setup()
    testDCLL.deleteHead()
    assert testDCLL.head.data == 1
    assert testDCLL.tail.next.data == 1
    testDCLL.deleteTail()
    assert testDCLL.tail.data != 4
    assert testDCLL.tail.next.data == 1
    testDCLL.delete(testDCLL.tail)
    testDCLL.delete(testNode6)
    testDCLL.insertHead(Node(11))
    testDCLL.insert(Node(11),3)
    testDCLL.insertTail(Node(11)) # insert 3 new duplicates
    testDCLL.delete(Node(11))
    currentTestNode = testDCLL.head
    testList = []
    for i in range(0, testDCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [1,2,7,10]
    testDCLL.clear()
    assert testDCLL.size == 0 and testDCLL.head == None and testDCLL.tail == None
        
@pytest.mark.test_search
def test_Search():
    testDCLL = setup()
    testNode6Search = testDCLL.search(testNode6)
    assert testNode6Search == testNode6
    nonListNode = Node(99)
    testNonListNode = testDCLL.search(nonListNode)
    assert testNonListNode == None

@pytest.mark.test_sorting
def test_Sorting():
    testDCLL = setup()
    testDCLL.sort()
    testList = []
    currentTestNode = testDCLL.head
    for i in range(0, testDCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList == [1,2,3,3,4,6,7,10]

    testIsSorted = testDCLL.isSorted()

    assert testIsSorted == True

    newTestNode = Node(8)
    duplicateTestNode = Node(4)
    # with patch.object(DLL, 'sort') as mock_sort:
    #         # call sortedInsert
    #         dLL.sortedInsert(newTestNode)

    #         # assert that the sort method was not called
    #         mock_sort.assert_not_called()

    testDCLL.sortedInsert(newTestNode)
    testDCLL.sortedInsert(duplicateTestNode)
    currentTestNode = testDCLL.head
    testList2 = []
    for i in range(0, testDCLL.size):
        testList2.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList2 == [1,2,3,3,4,4,6,7,8,10]


@pytest.mark.test_print
def test_Print():
    testDCLL = setup()
    captured_output = io.StringIO()
    sys.stdout = captured_output
    testDCLL.print()
    actual_output = captured_output.getvalue()
    sys.stdout = sys.stdout
    expected_output = "The size of this doubly circular linked list is 8.\nThis doubly circular linked list is not sorted.\nThis doubly circular linked list contains:\n6\n1\n2\n7\n10\n3\n3\n4\n"


import sys
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Linear.SinglyLL import singlyLL
from Nodes.SinglyNode import Node
class singlyCLL(singlyLL):
     
    def __init__(self, newNode = None):
        if newNode == None:   
            super().__init__()
        else:
            self.head = newNode
            self.head.next = self.head
            self.tail = newNode
            self.size = 1

    def insertHead(self, newNode):
        super().insertHead(newNode)
        self.tail.next = self.head
    
    def insertTail(self, newNode):
        super().insertTail(newNode)
        self.tail.next = self.head

    def insert(self, newNode, position):
        super().insert(newNode, position)
        if position == 1:
            self.tail.next = newNode
        if position == self.size:
            self.tail.next = self.head
             
            
    def sortedInsert(self, newNode): 
        isSorted = self.isSorted()
        if isSorted == False:
            self.sort()

        if self.size == 0:
            self.insertHead(newNode)

        # else:
        #     if newNode.data < self.head.data:
        #         self.insertHead(newNode) # need to use insert head and tail where the loop is created
        #     if newNode.data > self.tail.data:
        #         self.insertTail(newNode)
        #     else:
        #         current = self.head.next
        #         while current != self.tail:
        #             if newNode.data > current.data :          
        #                 newNode.next = current.next
        #                 current.next = newNode
        #                 self.size += 1
        #                 break
        elif newNode.data < self.head.data:
            self.insertHead(newNode)
        elif newNode.data > self.tail.data:
            self.insertTail(newNode)
        else:
            current_node = self.head
            while current_node.next is not self.head and current_node.next.data < newNode.data:
                current_node = current_node.next
            newNode.next = current_node.next
            current_node.next = newNode
            self.size += 1

        
        
                   
                
    def isSorted(self):
        return super().isSorted()

    def sort(self): # nope
        super().sort()
        self.tail.next = self.head

    def search(self, newNode):
        return(super().Search(newNode))

    def deleteHead(self):
        super().deleteHead()
        self.tail.next = self.head

    def deleteTail(self):
        super().deleteTail()
        self.tail.next = self.head

    def delete(self, delNode):
        tailModNeeded = False
        if delNode == self.head or delNode == self.tail:
            tailModNeeded = True
        super().delete(delNode)
        if tailModNeeded:
            self.tail.next = self.head
    
    def clear(self):
        super().clear()

    def print(self):
        print(f'The size of this singly circular linked list is {self.size}.')
        if self.isSorted():
            print("This singly circular linked list is sorted.")
        if not self.isSorted():
            print("This singly circular linked list is not sorted.")
        print("This singly circular linked list contains:")
        current = self.head
        for i in range(self.size):
            print(current.data)
            current = current.next
        

# def main():  # Delete later
#     testNode1 = Node(0)
#     testNode2 = Node(1)
#     testNode3 = Node(6)
#     testNode4 = Node(10)
#     testNode5 = Node(2)
#     testNode6 = Node(6)

#     testLL = singlyCLL()
#     testLL.insertTail(testNode1)
#     testLL.insertTail(testNode2)
#     testLL.insertTail(testNode3)
#     testLL.insertTail(testNode4)
#     testLL.insertTail(testNode6)
#     testLL.insertTail(testNode5)
#     testLL.delete(testNode6)
#     testLL.print()
# if  __name__  == "__main__":
#     main()




import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Linear.SinglyCLL import singlyCLL
from datastructures.Nodes.SinglyNode import Node

testNode1 = Node(6)
testNode2 = Node(1)
testNode3 = Node(2)
testNode4 = Node(7)
testNode5 = Node(10)   
testNode6 = Node(3)
testNode6_2 = Node(3)
testNode7 = Node(4)

@pytest.mark.test_constructor
def test_constructor():
    testSCLL = singlyCLL()
    assert testSCLL.head == None
    assert testSCLL.tail == None
    
    testSCLL2 = singlyCLL(testNode1)
    assert testSCLL2.head.data == 6
    assert testSCLL2.tail.data == 6
    assert testSCLL2.size == 1
    assert testSCLL2.tail.next.data == 6

@pytest.mark.test_insert
def test_insert():
    testSCLL = singlyCLL()
    testSCLL.insertHead(testNode1)
    assert testSCLL.head.next.data == 6 # Added this additional test to ensure circular with one element
    assert testSCLL.head.data == 6
    testSCLL.insertHead(testNode2)
    testSCLL.insertTail(testNode3) 
    testSCLL.insertTail(testNode4)
    assert testSCLL.tail.data == 7
    testSCLL.insert(testNode5, 3)
    assert testSCLL.head.next.next.data == 10
    assert testSCLL.tail.next.data == 1 # Added this additional test to ensure circular with multiple elements
    testSCLL.insert(testNode6,5)
    testSCLL.insert(Node(11), 7)  # Added this additional test
    assert testSCLL.tail.data == 11 # Added this additional test
    testSCLL.insert(testNode7,1) # test insert at head
    currentTestNode = testSCLL.head 
    testList = []
    for i in range(0, testSCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4,1,6,10,2,3,7,11]
    testSCLL.insertTail(testNode6_2)
    currentTestNode = testSCLL.head 
    testList = []
    for i in range(0, testSCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4,1,6,10,2,3,7,11,3]
    assert testSCLL.tail.next.data == 4 # Added this additional test to ensure circular with multiple elements


def setup():
    testSCLL = singlyCLL()
    testSCLL.insertTail(testNode1)
    testSCLL.insertTail(testNode2)
    testSCLL.insertTail(testNode3) 
    testSCLL.insertTail(testNode4)
    testSCLL.insertTail(testNode5)
    testSCLL.insertTail(testNode6)
    testSCLL.insertTail(testNode6_2)
    testSCLL.insertTail(testNode7)
    return testSCLL

@pytest.mark.test_isCircular
def test_isCircular():
    testSCLL = setup()
    circTestNode = testSCLL.tail
    assert testSCLL.head.data == circTestNode.next.data

@pytest.mark.test_delete
def test_delete():
    testSCLL = setup()
    testSCLL.deleteHead()
    assert testSCLL.head.data != 6
    assert testSCLL.tail.next.data == 1 # Added this additional test
    testSCLL.deleteTail()
    assert testSCLL.tail.data != 4
    assert testSCLL.tail.next.data == 1 # Added this additional test
    testSCLL.delete(testSCLL.tail)
    testSCLL.delete(testNode6)
    #*********************************
    # These 4 lines are new
    testSCLL.insertHead(Node(11))
    testSCLL.insert(Node(11), 3)
    testSCLL.insertTail(Node(11)) # insert 3 new duplicates
    testSCLL.delete(Node(11))
    currentTestNode = testSCLL.head
    testList = []
    for i in range(0, testSCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [1,2,7,10]
    testSCLL.clear()
    assert testSCLL.size == 0 and testSCLL.head == None and testSCLL.tail == None

@pytest.mark.test_search
def test_Search():
    testDCLL = setup()
    testNode6Search = testDCLL.search(testNode6)
    assert testNode6Search == testNode6
    nonListNode = Node(99)
    testNonListNode = testDCLL.search(nonListNode)
    assert testNonListNode == None

@pytest.mark.test_sorting
def test_Sorting():
    testSCLL = setup()
    testSCLL.sort()
    testList = []
    currentTestNode = testSCLL.head
    for i in range(0, testSCLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList == [1,2,3,3,4,6,7,10]

    testIsSorted = testSCLL.isSorted()

    assert testIsSorted == True

    newTestNode = Node(8)
    duplicateTestNode = Node(4)

    testSCLL.sortedInsert(newTestNode)
    testSCLL.sortedInsert(duplicateTestNode)
    currentTestNode = testSCLL.head
    testList2 = []
    for i in range(0, testSCLL.size):
        testList2.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList2 == [1,2,3,3,4,4,6,7,8,10]

    testSCLL2 = setup()
    testSCLL2.sortedInsert(Node(11))
    testSCLL2.sortedInsert(Node(0))
    current = testSCLL2.head
    testList3 = []
    for i in range(0, testSCLL2.size):
        testList3.append(current.data)
        current = current.next
    assert testList3 == [0,1,2,3,3,4,6,7,10,11]

    testSCLL3 = singlyCLL()
    testSCLL3.sortedInsert(Node(4))
    assert testSCLL3.head.data == 4
    assert testSCLL3.tail.data == 4


@pytest.mark.test_print
def test_Print():
    testSCLL = setup()
    captured_output = io.StringIO()
    sys.stdout = captured_output
    testSCLL.print()
    actual_output = captured_output.getvalue()
    sys.stdout = sys.stdout
    expected_output = "The size of this singly circular linked list is 8.\nThis singly circular linked list is not sorted.\nThis singly circular linked list contains:\n6\n1\n2\n7\n10\n3\n3\n4\n"



import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Nodes.TNode import TNode
from datastructures.Trees.AVL import AVL
from datastructures.Trees.BST import BST

def is_avl_balanced(node):
    """
    Check if the given AVL tree node is balanced, i.e. all nodes to the left have values less than
    the node's value, and all nodes to the right have values greater than the node's value. Also
    checks that the tree is balanced according to the node's balance factor.
    """
    if node is None:
        return True

    if node.balance not in [-1, 0, 1]:
        # the node is unbalanced
        return False

    if node.leftChild and node.leftChild.data > node.data:
        # the left subtree contains a node with a value greater than or equal to the current node
        return False

    if node.rightChild and node.rightChild.data < node.data:
        # the right subtree contains a node with a value less than or equal to the current node
        return False

    # recursively check the subtrees
    return is_avl_balanced(node.leftChild) and is_avl_balanced(node.rightChild)

def inorderTraversal(node):
    values = []
    if node is not None:
        values += inorderTraversal(node.leftChild)
        values.append(node.data)
        values += inorderTraversal(node.rightChild)
    return values


@pytest.mark.test_constructor
def test_constructor():
    AVL1 = AVL()
    assert AVL1.root == None

    AVL2 = AVL(TNode(4))
    assert AVL2.root.data == 4
    assert AVL2.root.parent == None
    assert AVL2.root.leftChild == None
    assert AVL2.root.rightChild == None
    assert AVL2.root.balance == 0


    AVL3 = AVL(4)
    assert AVL3.root.data == 4
    assert AVL3.root.parent == None
    assert AVL3.root.leftChild == None
    assert AVL3.root.rightChild == None
    assert AVL3.root.balance == 0



@pytest.mark.test_insert
def test_insert():
    testAVL = AVL(TNode(6))
    testAVL.insert(TNode(4))
    testAVL.insert(9)
    testAVL.insert(2)
    testAVL.insert(5)
    testAVL.insert(TNode(7))
    testAVL.insert(TNode(3))
    testAVL.insert(1)
    testAVL.insert(0)
    currentTestNode = testAVL.root
    assert currentTestNode.data == 6
    assert currentTestNode.leftChild.data == 2
    assert currentTestNode.rightChild.data == 9
    assert currentTestNode.leftChild.leftChild.data == 1
    assert currentTestNode.leftChild.rightChild.data == 4
    assert currentTestNode.rightChild.leftChild.data == 7
    assert currentTestNode.leftChild.leftChild.leftChild.data == 0
    assert currentTestNode.leftChild.rightChild.rightChild.data == 5
    assert currentTestNode.leftChild.rightChild.leftChild.data == 3

    # assert is_avl_balanced(testAVL.root) == True

    # Case 1:
    AVL2 = AVL(TNode(4))
    AVL2.insert(TNode(2))
    AVL2.insert(TNode(7))
    AVL2.insert(1)
    AVL2.insert(6)
    AVL2.insert(9)
    AVL2.insert(8)
    assert AVL2.root.data == 4
    assert AVL2.root.leftChild.data == 2
    assert AVL2.root.rightChild.data == 7
    assert AVL2.root.leftChild.leftChild.data == 1
    assert AVL2.root.rightChild.leftChild.data == 6
    assert AVL2.root.rightChild.rightChild.data == 9
    assert AVL2.root.rightChild.rightChild.leftChild.data == 8

    # assert is_avl_balanced(AVL2.root) == True

    # Case 2
    AVL3 = AVL(TNode(4))
    AVL3.insert(TNode(2))
    AVL3.insert(TNode(7))
    AVL3.insert(1)
    AVL3.insert(6)
    AVL3.insert(9)
    AVL3.insert(8)
    AVL3.insert(10)
    assert AVL3.root.data == 4
    assert AVL3.root.leftChild.data == 2
    assert AVL3.root.rightChild.data == 7
    assert AVL3.root.leftChild.leftChild.data == 1
    assert AVL3.root.rightChild.leftChild.data == 6
    assert AVL3.root.rightChild.rightChild.data == 9
    assert AVL3.root.rightChild.rightChild.leftChild.data == 8
    assert AVL3.root.rightChild.rightChild.rightChild.data == 10

    # assert is_avl_balanced(AVL3.root) == True

    # Case 3a
    AVL4 = AVL(60)
    AVL4.insert(40)
    AVL4.insert(80)
    AVL4.insert(20)
    AVL4.insert(50)
    AVL4.insert(95)
    AVL4.insert(10)
    AVL4.insert(30)
    AVL4.insert(5)
    assert AVL4.root.data == 60
    assert AVL4.root.leftChild.data == 20
    assert AVL4.root.rightChild.data == 80
    assert AVL4.root.leftChild.leftChild.data == 10
    assert AVL4.root.leftChild.rightChild.data == 40
    assert AVL4.root.rightChild.rightChild.data == 95
    assert AVL4.root.leftChild.leftChild.leftChild.data == 5
    assert AVL4.root.leftChild.rightChild.rightChild.data == 50
    assert AVL4.root.leftChild.rightChild.leftChild.data == 30

    # assert is_avl_balanced(AVL4.root) == True

    # Case 3b
    AVL5 = AVL(TNode(2))
    AVL5.insert(TNode(1))
    AVL5.insert(6)
    AVL5.insert(4)
    AVL5.insert(7)
    AVL5.insert(5)
    assert AVL5.root.data == 4
    assert AVL5.root.leftChild.data == 2
    assert AVL5.root.rightChild.data == 6
    assert AVL5.root.leftChild.leftChild.data == 1
    assert AVL5.root.rightChild.rightChild.data == 7
    assert AVL5.root.rightChild.leftChild.data == 5

    # assert is_avl_balanced(AVL5.root) == True

    # Case with repeats
    AVL6 = AVL(TNode(4))
    AVL6.insert(TNode(2))
    AVL6.insert(7)
    AVL6.insert(2)
    AVL6.insert(2)
    assert AVL6.root.data == 4
    assert AVL6.root.leftChild.data == 2
    assert AVL6.root.rightChild.data == 7
    assert AVL6.root.leftChild.leftChild.data == 2
    assert AVL6.root.leftChild.rightChild.data == 2

    

@pytest.mark.test_setAndGetRoot
def test_setAndGetRoot():
    testAVL = AVL()
    testAVL.setRoot(2)
    assert testAVL.getRoot().data == 2

    AVL2 = AVL(TNode(4))
    AVL2.insert(TNode(2))
    AVL2.insert(TNode(7))
    AVL2.insert(1)
    AVL2.insert(6)
    AVL2.insert(9)
    AVL2.insert(8)
    assert AVL2.getRoot().data == 4
    AVL2.setRoot(5)
    assert AVL2.getRoot().data == 4
    AVL2.setRoot(0)
    assert is_avl_balanced(AVL2.root) == True

@pytest.mark.test_insertNode_hasChildren
def test_insertNode_hasChildren():
        node1 = TNode(2)
        node2 = TNode(1)
        node3 = TNode(6)
        node4 = TNode(4)
        node5 = TNode(7)
        node6 = TNode(5)
        AVL1 = AVL(node1)
        AVL1.insert(node2)
        AVL1.insert(node3)
        AVL1.insert(node4)
        AVL1.insert(node5)
        AVL1.insert(node6)
        node7 = TNode(8) 
        AVL2 = AVL(node7)
        AVL2.insert(TNode(3))
        AVL2.insert(TNode(12))
        AVL1.setRoot(node7)
        assert AVL1.root.data == 4
        assert AVL1.root.leftChild.data == 2
        assert AVL1.root.leftChild.leftChild.data == 1
        assert AVL1.root.leftChild.rightChild.data == 3
        assert AVL1.root.rightChild.data == 8
        assert AVL1.root.rightChild.rightChild.data == 12
        assert AVL1.root.rightChild.leftChild.data == 6
        assert AVL1.root.rightChild.leftChild.leftChild.data == 5
        assert AVL1.root.rightChild.leftChild.rightChild.data == 7
        Node8 = TNode(2)
        testBST = BST(Node8)
        testBST.insert(1)
        testBST.insert(6)
        testBST.insert(4)
        testBST.insert(7)
        testBST.insert(5)
        AVL3 = AVL(Node8)
        assert AVL3.root.data == 4
        assert AVL3.root.leftChild.data == 2
        assert AVL3.root.leftChild.leftChild.data == 1
        assert AVL3.root.rightChild.data == 6
        assert AVL3.root.rightChild.rightChild.data == 7
        assert AVL3.root.rightChild.leftChild.data == 5

@pytest.mark.test_delete
def test_delete():
    AVL1 = AVL(TNode(4))
    AVL1.insert(TNode(2))
    AVL1.insert(TNode(7))
    AVL1.insert(1)
    AVL1.insert(6)
    AVL1.insert(9)
    AVL1.insert(8)
    AVL1.delete(1)
    assert AVL1.root.data == 7
    assert AVL1.root.leftChild.data == 4
    assert AVL1.root.rightChild.data == 9
    assert AVL1.root.leftChild.leftChild.data == 2
    assert AVL1.root.rightChild.leftChild.data == 8
    assert AVL1.root.leftChild.rightChild.data == 6

    assert is_avl_balanced(AVL1.root) == True

    AVL1.delete(8)
    assert AVL1.root.data == 7
    assert AVL1.root.leftChild.data == 4
    assert AVL1.root.rightChild.data == 9
    assert AVL1.root.leftChild.leftChild.data == 2
    assert AVL1.root.rightChild.leftChild == None
    assert AVL1.root.leftChild.rightChild.data == 6

    assert is_avl_balanced(AVL1.root) == True   

    # Testing deleting duplicates
    AVL6 = AVL(TNode(4))
    AVL6.insert(TNode(2))
    AVL6.insert(7)
    AVL6.insert(2)
    AVL6.insert(2)
    AVL6.delete(2)
    nodes = inorderTraversal(AVL6.root)
    assert (2 not in nodes)

@pytest.mark.test_search
def test_search():
    AVL1 = AVL(TNode(4))
    AVL1.insert(TNode(2))
    AVL1.insert(TNode(7))
    AVL1.insert(1)
    AVL1.insert(6)
    AVL1.insert(9)
    AVL1.insert(8)
    assert AVL1.search(1).data == 1
    assert AVL1.search(4).data == 4
    assert AVL1.search(2).data == 2
    assert AVL1.search(7).data == 7
    assert AVL1.search(6).data == 6
    assert AVL1.search(9).data == 9
    assert AVL1.search(8).data == 8

@pytest.mark.test_print
def test_print():
    AVL1 = AVL(TNode(4))
    AVL1.insert(TNode(2))
    AVL1.insert(TNode(7))
    AVL1.insert(1)
    AVL1.insert(6)
    AVL1.insert(9)
    AVL1.insert(8)

    capturedOutput = io.StringIO()
    sys.stdout = capturedOutput
    AVL1.printInOrder()
    actual_output = capturedOutput.getvalue()
    sys.stdout = sys.__stdout__
    expected_output = "1\n2\n4\n6\n7\n8\n9\n"
    assert actual_output == expected_output

    actual_output = ""
    capturedOutput = io.StringIO()
    sys.stdout = capturedOutput
    AVL1.printBF()
    actual_output = capturedOutput.getvalue()
    expected_output = "4 \n2 7 \n1 6 9 \n8 \n"
    assert expected_output == actual_output

    


import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Nodes.TNode import TNode
from datastructures.Trees.BST import BST

testNode1 = TNode(6)
testNode2 = TNode(4)
testNode3 = TNode(9)
testNode4 = TNode(2)
testNode5 = TNode(5)   
testNode6 = TNode(7)
testNode7 = TNode(3)


def is_bst_balanced(node):
    """
    Check if the given AVL tree node is balanced, i.e. all nodes to the left have values less than
    the node's value, and all nodes to the right have values greater than the node's value. Also
    checks that the tree is balanced according to the node's balance factor.
    """
    if node is None:
        return True

    if node.leftChild and node.leftChild.data > node.data:
        # the left subtree contains a node with a value greater than or equal to the current node
        return False

    if node.rightChild and node.rightChild.data < node.data:
        # the right subtree contains a node with a value less than or equal to the current node
        return False

    # recursively check the subtrees
    return is_bst_balanced(node.leftChild) and is_bst_balanced(node.rightChild)

def inorderTraversal(node):
    values = []
    if node is not None:
        values += inorderTraversal(node.leftChild)
        values.append(node.data)
        values += inorderTraversal(node.rightChild)
    return values

@pytest.mark.test_constructor
def test_constructor():
    testBST = BST()
    assert testBST.root == None

    testBST2 = BST(TNode(2))
    assert testBST2.root.data == 2
    assert testBST2.root.parent == None
    assert testBST2.root.leftChild == None
    assert testBST2.root.rightChild == None
    assert testBST2.root.balance == None

    testBST2 = BST(2)
    assert testBST2.root.data == 2
    assert testBST2.root.parent == None
    assert testBST2.root.leftChild == None
    assert testBST2.root.rightChild == None
    assert testBST2.root.balance == None

@pytest.mark.test_insert
def test_insert():
    testBST = BST(testNode1)
    testBST.insert(testNode2)
    testBST.insert(9)
    testBST.insert(2)
    testBST.insert(5)
    testBST.insert(testNode6)
    testBST.insert(testNode7)
    currentTestNode = testBST.root
    assert currentTestNode.data == 6
    assert currentTestNode.leftChild.data == 4
    assert currentTestNode.rightChild.data == 9
    assert currentTestNode.leftChild.leftChild.data == 2
    assert currentTestNode.leftChild.rightChild.data == 5
    assert currentTestNode.rightChild.leftChild.data == 7
    assert currentTestNode.leftChild.leftChild.rightChild.data == 3
    testBST.insert(2)
    testBST.insert(2)
    assert is_bst_balanced(testBST.root) == True

def setup():
    testBST = BST(6)
    testBST.insert(4)
    testBST.insert(9)
    testBST.insert(2)
    testBST.insert(5)
    testBST.insert(7)
    testBST.insert(3)
    return testBST

@pytest.mark.test_setAndGetRoot
def test_setAndGetRoot(): # does this new root need old root as a child along with it's children?
    testBST = setup()
    assert testBST.getRoot().data == 6
    testBST.setRoot(8)
    assert testBST.getRoot().data == 8
    assert testBST.getRoot().leftChild.data < 8
    assert testBST.getRoot().rightChild.data > 8
    assert is_bst_balanced(testBST.root) == True

@pytest.mark.test_delete
def test_delete():
    testBST = setup()
    testBST.insert(2)
    testBST.insert(2)
    testBST.delete(2)
    testBST.delete(6)
    testBST.delete(5)
    testBST.delete(4)
    testList = inorderTraversal(testBST.getRoot())
    assert (6 not in testList and 2 not in testList and 5 not in testList and 4 not in testList)
    assert is_bst_balanced(testBST.root) == True
    
@pytest.mark.test_search
def test_search():
    testBST = setup()
    assert testBST.search(2).data == 2
    assert testBST.search(6).data == 6
    assert testBST.search(5).data == 5
    assert testBST.search(4).data == 4
    assert testBST.search(1) == None
    assert testBST.search(8) == None
    assert testBST.search(10) == None

@pytest.mark.test_print
def test_print():
    testBST = setup()
    capturedOutput = io.StringIO()
    sys.stdout = capturedOutput
    testBST.printInOrder()
    actual_output = capturedOutput.getvalue()
    sys.stdout = sys.__stdout__
    expected_output = "2\n3\n4\n5\n6\n7\n9\n"
    assert actual_output == expected_output
    
    actual_output = ""
    capturedOutput = io.StringIO()
    sys.stdout = capturedOutput
    testBST.printBF()
    actual_output = capturedOutput.getvalue()
    expected_output = "6 \n4 9 \n2 5 7 \n3 \n"
    assert expected_output == actual_output




import sys
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Trees.BST import BST
from Nodes.TNode import TNode

class AVL(BST):
    def __init__(self, arg = None):
        if arg == None:
            super().__init__()
        else:
            super().__init__(arg) # sets newNode as root 
            if self.root.leftChild is not None or self.root.rightChild is not None:
                node_list = self.inorder_traversal()
                root = node_list[0]
                self.root = root
                root.balance = 0
                root.leftChild = None
                root.rightChild = None
                node_list.remove(root)
                for node in node_list:
                        node.parent = None
                        node.leftChild = None
                        node.rightChild = None
                        node.balance = 0
                        self.insert(node) 
    
    def inorder_traversal(self, node=None, node_list=None):
        if node_list is None:
            node_list = []
        if self.getRoot() is None:
            return
        if node is None:
            node = self.getRoot()
        queue = [node]
        while queue:
            curr_node = queue.pop(0)
            left_height = self.get_subtree_height(curr_node.leftChild)
            right_height = self.get_subtree_height(curr_node.rightChild)
            curr_node.balance = right_height - left_height
            node_list.append(curr_node)
            if curr_node.leftChild:
                queue.append(curr_node.leftChild)
            if curr_node.rightChild:
                queue.append(curr_node.rightChild)
        return node_list

    def get_subtree_height(self, node=None):
        if node is None:
            return -1
        left_height = self.get_subtree_height(node.leftChild)
        right_height = self.get_subtree_height(node.rightChild)
        return 1 + max(left_height, right_height)

    def balance_tree(self, insert):
        current = self.getRoot()
        pivot = current
        while current is not None:
            if current.balance > 0 or current.balance < 0:
                pivot = current
            if insert.data > current.data: # go right
                current = current.rightChild
            else:
                current = current.leftChild # go left

        if pivot.balance == 1 or pivot.balance == -1:
            if insert.data > pivot.data:
                son = pivot.rightChild
            else:
                son = pivot.leftChild
            ancestor = pivot.parent

            if pivot.balance == 1 and insert.data < pivot.data or pivot.balance == -1 and insert.data > pivot.data: # case 2:
                self.inorder_traversal()
                return
            
            insert = self.search(insert.data)

            if pivot.balance < 0 and insert.data <= son.data: # case 3a 
                self.rightRotate(pivot,son, ancestor, insert)
            
            elif pivot.balance > 0 and insert.data> son.data:  # case 3a
                self.leftRotate(pivot, son, ancestor, insert)
            
            else:
                if insert.data > son.data:
                    grandson = son.rightChild
                else:   
                    grandson = son.leftChild
                if pivot.balance < 0 and insert.data > son.data: # case 3b
                    self.leftRightRotate(pivot, son, grandson, ancestor,insert)
                
                elif pivot.balance > 0 and insert.data < son.data: # case 3b
                    self.rightLeftRotate(pivot, son, grandson,ancestor, insert)
        else: #case 1
            self.inorder_traversal()


    def rightRotate(self, pivot, son, ancestor, insert):
        if ancestor is not None:
            if ancestor.data > son.data:
                ancestor.leftChild = son
            else:
                ancestor.rightChild = son
            son.parent = ancestor
        pivot.leftChild = son.rightChild
        if son.rightChild is not None:
            son.rightChild.parent = pivot

        if pivot == self.getRoot():
            # pivot.leftChild = son.rightChild
            self.root = son
            son.parent = None
        son.rightChild = pivot
        pivot.parent = son
        
        #adjust balances
        pivot.balance = 0
        insert.balance = 0


    def leftRotate(self, pivot, son, ancestor, insert):
        if ancestor is not None:
            if ancestor.data < son.data:
                ancestor.rightChild = son
            else:
                ancestor.leftChild = son
            son.parent = ancestor

        pivot.rightChild = son.leftChild
        if son.leftChild is not None:
            son.leftChild.parent = pivot

        if pivot == self.getRoot():
            # pivot.rightChild = son.leftChild
            self.root = son
            son.parent = None
        son.leftChild = pivot
        pivot.parent = son
        
        #adjust balances
        pivot.balance = 0
        insert.balance = 0

    def rightLeftRotate(self, pivot, son, grandson, ancestor, insert):
        self.rightRotate(son, grandson, pivot, insert)
        self.leftRotate(pivot, grandson, ancestor, insert)
        if insert.balance > grandson.balance:
            pivot.balance = -1
        else:
            pivot.balance = 0
            son.balance = 1
        
        temp = insert
        while (temp != pivot.leftChild and temp != pivot.rightChild  and temp != son.leftChild and temp != son.rightChild) and grandson != insert:
            right = self.get_subtree_height(temp.rightChild)
            left = self.get_subtree_height(temp.leftChild)
            temp.balance = right - left
            temp = temp.parent

    def leftRightRotate(self, pivot, son, grandson,ancestor, insert):
        self.leftRotate(son,grandson,pivot, insert)
        self.rightRotate(pivot, grandson, ancestor, insert)
        if insert.balance < grandson.balance:
            pivot.balance = 1
        else:
            pivot.balance = 0
            son.balance = -1

        temp = insert
        while (temp != pivot.leftChild and temp != pivot.rightChild  and temp != son.leftChild and temp != son.rightChild) and grandson != insert:
            right = self.get_subtree_height(temp.rightChild)
            left = self.get_subtree_height(temp.leftChild)
            temp.balance = right - left
            temp = temp.parent

    def insert(self, arg):
        if isinstance(arg, int):
            node = TNode(arg)
        else:
            node = arg
        self.inorder_traversal()
        super().insert(node)
        self.balance_tree(node)
        
    def setRoot(self, newNode):
        if isinstance(newNode, int):
            newRoot = TNode(newNode)
        else:
            newRoot = newNode
        node_list = self.inorder_traversal() # store prexisting tree
        if newRoot.leftChild is not None or newRoot.rightChild is not None:
            self.root = newRoot
            node_list1 = self.inorder_traversal()
            root = node_list1[0]
            node_list1.remove(root)
            root.balance = 0
            root.leftChild = None
            root.rightChild = None
            for node in node_list1:
                    node.parent = None
                    node.leftChild = None
                    node.rightChild = None
                    node.balance = 0
                    self.insert(node) 

        if node_list is not None: 
            self.root = newRoot
            for node in node_list:
                node.parent = None
                node.leftChild = None
                node.rightChild = None
                node.balance = 0
                self.insert(node)    
        else:
            self.root = newRoot

    def getRoot(self):
        return super().getRoot()
            
    def delete(self, data):
        super().delete(data)
        node_list = self.inorder_traversal()
        root = node_list[0]
        self.root = root
        root.balance = 0
        root.leftChild = None
        root.rightChild = None
        node_list.remove(root)
        for node in node_list:
                node.parent = None
                node.leftChild = None
                node.rightChild = None
                node.balance = 0
                self.insert(node) 
    
    def search(self, data):
        return super().search(data)
    
    def printInOrder(self):
        super().printInOrder()
    
    def printBF(self):
        super().printBF()


from pathlib import Path
import sys
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Nodes.DoublyNode import Node
from Linear.DoublyLL import doublyLL

class doublyCLL(doublyLL):
    # Note that you can overload by including the exact same def statement (exact same method name, with same arguments) and when it is called on an instance of the child class, the override method will be called
    def __init__(self, newNode = None):
        self.head = newNode
        self.tail = newNode
        if newNode == None:
            self.size = 0
        else:
            self.head.next = self.head
            self.head.prev = self.head
            self.size = 1
            
    def insertHead(self, newNode):
        super().insertHead(newNode)
        self.tail.next = self.head
        self.head.prev = self.tail
        

    def insertTail(self, newNode):
        super().insertTail(newNode)
        self.tail.next = self.head
        self.head.prev = self.tail


    def insert(self, newNode, position):
        super().insert(newNode, position)
        if position == 1 or position == self.size:
            self.tail.next = self.head
            self.head.prev = self.tail

    def sortedInsert(self, newNode): # INCOMPLETE IMPLEMENTATION
        sortNeeded = self.isSorted()
        if sortNeeded == False:
            self.sort()

        if self.size == 0:
            self.insertHead(newNode)
        elif newNode.data < self.head.data:
            self.insertHead(newNode)
        elif newNode.data > self.tail.data:
            self.insertTail(newNode)
        else:
            current_node = self.head
            while current_node.next is not self.head and current_node.next.data < newNode.data:
                current_node = current_node.next
            newNode.prev = current_node
            newNode.next = current_node.next
            current_node.next.prev = newNode
            current_node.next = newNode
            self.size += 1


    def isSorted(self):
        return super().isSorted()

    def sort(self):
        i = self.head.next
        j = self.head
        for a in range(1, self.size): # a is some random variable to set the loop iterations
            key = i.data
            while j != self.tail and key < j.data:
                j.next.data = j.data
                j = j.prev
            j.next.data = key
            i = i.next
            j = i.prev

    def search(self, newNode):
        return(super().search(newNode))

    def deleteHead(self):
        super().deleteHead()
        self.tail.next = self.head
        self.head.prev = self.tail

    def deleteTail(self):
        super().deleteTail()
        self.tail.next = self.head
        self.head.prev = self.tail

    def delete(self, delNode):
        
        super().delete(delNode)
        
        self.tail.next = self.head
        self.head.prev = self.tail

    def clear(self):
        super().clear()

    def print(self):
        print(f'The size of this doubly circular linked list is {self.size}.')
        if self.isSorted():
            print("This doubly circular linked list is sorted.")
        if not self.isSorted():
            print("This doubly circular linked list is not sorted.")
        print("This doubly circular linked list contains:")
        current = self.head
        for i in range(self.size):
            print(current.data)
            current = current.next




# def main():
#     test1 = doublyCLL()
#     test1.insertHead(Node(1))
#     test1.insertHead(Node(2))
#     test1.insertTail(Node(11))
#     test1.insertTail(Node(7))
#     test1.insertTail(Node(3))
#     test1.insertTail(Node(60))
#     test1.insertTail(Node(4))
#     print(test1.search(Node(4)))
#     test1.sort()
#     test1.print()

#     # test2 = doublyCLL()
#     # test2.insertHead(Node(4))
#     # test2.insertHead(Node(3))
#     # test2.insertHead(Node(2))
#     # test2.sort()
#     # test2.print()

# if __name__ == "__main__":
#     main()

    
    



import sys
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Linear.SinglyLL import singlyLL
from Nodes.SinglyNode import Node

class LLQueue(singlyLL):

    def __init__(self, newNode = None):
        super().__init__(newNode)

    def enqueue(self, newNode):
        super().insertTail(newNode)

    def dequeue(self):
        super().deleteHead()

    def clear(self):
        super().clear()
    
    # def sortedEneque(self, newNode): 
    #     sortNeeded = self.isSorted()
    #     if sortNeeded == False:
    #         self.enqueue(newNode)
    #         self.sort()
    #     else:
    #         self.enqueue(newNode)
       
            
    def isSorted(self):
        return

    def sort(self):
        return

    def search(self, newNode):
        return super().search(newNode)


    def print(self):
        print(f'The size of this singly linked list queue is {self.size}.')
        if self.isSorted():
            print("This singly linked list queue is sorted.")
        if not self.isSorted():
            print("This singly linked list queue is not sorted.")
        print("This singly linked list queue contains:")
        current = self.head
        while current != None:
            print(current.data)
            current = current.next

    def insertHead(self, newNode):
        return
    
    def insert(self, newNode, position):
        return 
    
    def delete(self, delNode):
        return 
    
    def deleteTail(self):
        return 
    
    def deleteHead(self):
        return
    
    def insertTail(self, newNode):
        return
    
    def sortedInsert(self, newNode):
        return
        
# def main():  # Delete later
#     testNode1 = Node(6)
#     testNode2 = Node(1)
#     testNode3 = Node(2)
#     testNode4 = Node(7)
#     testNode5 = Node(10)
#     testQueue = LLQueue(testNode1)
#     testQueue.enqueue(testNode2)
#     testQueue.enqueue(testNode3)
#     testQueue.enqueue(testNode4)
#     testQueue.enqueue(testNode5)
#     # testLL.sort()
#     testQueue.print()


# if  __name__  == "__main__":
#     main()


            



   
        





import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Linear.LLStack import LLStack
from datastructures.Nodes.SinglyNode import Node

testNode1 = Node(6)
testNode2 = Node(1)
testNode3 = Node(2)
testNode4 = Node(7)
testNode5 = Node(10)   
testNode6 = Node(3)
testNode6_2 = Node(3)
testNode7 = Node(4)

@pytest.mark.test_constructor
def test_constructor():
    testStack = LLStack()
    assert testStack.head == None
    assert testStack.tail == None
    testStack2 = LLStack(Node(2))
    assert testStack2.head.data == 2
    assert testStack2.tail.data == 2

@pytest.mark.test_push
def test_push():
    testStack = LLStack()
    testStack.push(testNode1)
    assert testStack.head.data == 6
    testStack.push(testNode2)
    testStack.push(testNode3) 
    testStack.push(testNode4)
    testStack.push(testNode5)
    assert testStack.head.next.next.data == 2
    testStack.push(testNode6)
    testStack.push(testNode7) # test push at head
    currentTestNode = testStack.head 
    testList = []
    for i in range(0, testStack.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4,3,10,7,2,1,6]
    testStack.push(testNode6_2)
    testList = []
    currentTestNode = testStack.head
    for i in range(0, testStack.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList == [3,4,3,10,7,2,1,6]
    
def setup():
    testStack = LLStack()
    testStack.push(testNode1)
    testStack.push(testNode2)
    testStack.push(testNode3) 
    testStack.push(testNode4)
    testStack.push(testNode5)
    testStack.push(testNode6)
    testStack.push(testNode6_2)
    testStack.push(testNode7)
    return testStack

@pytest.mark.test_pop
def test_pop():
    testStack = setup()
    testStack.pop()
    assert testStack.head.data != 6
    testStack.pop()
    testStack.pop()
    currentTestNode = testStack.head
    testList = []
    for i in range(0, testStack.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [10,7,2,1,6]
    testStack.clear()
    assert testStack.size == 0 and testStack.head == None
        
@pytest.mark.test_search
def test_search():
    testStack = setup()
    testNode6Search = testStack.search(testNode7)
    assert testNode6Search == testNode7
    nonListNode = Node(99)
    testNonListNode = testStack.search(nonListNode)
    assert testNonListNode == None

@pytest.mark.test_print
def test_print():
    testStack = setup()
    captured_output = io.StringIO()
    sys.stdout = captured_output
    testStack.print()
    actual_output = captured_output.getvalue()

    sys.stdout = sys.__stdout__

    expected_output = "The size of this singly linked list stack is 8.\nThis singly linked list stack is not sorted.\nThis singly linked list stack contains:\n4\n3\n3\n10\n7\n2\n1\n6\n"
    assert actual_output ==  expected_output

@pytest.mark.test_detreimentFunctionality
def test_detrimentFunctionality():
    testStack = LLStack(testNode7)
    testStack.insertHead(testNode1)
    testStack.insert(testNode2,1)
    testStack.insertTail(testNode3)
    testStack.deleteHead()
    testStack.deleteTail()
    testStack.delete(testNode7)
    testList = []
    currentTestNode = testStack.head
    for i in range(testStack.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4]
    testStack.push(Node(6))
    testStack.sort()
    testStack.sortedInsert(Node(8))

    testList = []
    currentTestNode = testStack.head
    for i in range(testStack.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [6, 4]
    



import sys
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Nodes.DoublyNode import Node

# This class does not extend the SLL class as it requires too much rewriting.

class doublyLL:
    def __init__(self, newNode = None):
        self.head = newNode
        self.tail = newNode
        if newNode == None:
            self.size = 0
        else:
            self.size = 1

    # InsertHead(node): Inserts node at the head of the list
    def insertHead(self, newNode):
        if self.head is None:
            self.head = newNode
            self.tail = newNode
        else:
            self.head.prev = newNode
            newNode.next = self.head
            self.head = newNode
        self.size += 1


    # InsertTail(node): Inserts node at the tail of the list
    def insertTail(self, newNode):
        if self.head is None:
            self.head = newNode
            self.tail = newNode
        else:
            self.tail.next = newNode
            self.tail.next.prev = self.tail
            self.tail = self.tail.next
        self.size += 1

    # Insert(node, position): Inserts node a the specified position
    def insert(self, newNode, position): # DOUBLE CHECK: ASSUMING THAT THE POSITION STARTS FROM ZERO AND NOT ONE
        if position < 1 or position > (self.size + 1):
            print("Error. Out of range. Operation exited.")
        else:
            if self.head is None:
                self.head = newNode
                self.tail = newNode
            elif position == 1:
                self.head.prev = newNode
                newNode.next = self.head
                self.head = newNode
            elif position == self.size + 1:
                self.tail.next = newNode
                self.tail.next.prev = self.tail
                self.tail = self.tail.next
            
            else: # Insert at a specific position
                current_node = self.head
                for i in range(position - 2):
                    current_node = current_node.next

                newNode.next = current_node.next
                current_node.next.prev = newNode
                current_node.next = newNode
                newNode.prev = current_node

            self.size += 1

    # isSorted(self): Checks the current DLL to see if it is already sorted
    def isSorted(self): # DOUBLE CHECK: HOW DO WE KNOW THAT IT IS SORTED IN THIS ORDER?
        if self.head is None or self.head.next is None:
            return True
        current = self.head
        for i in range(self.size-1):
            if current.next.data < current.data:
                return False
            current = current.next
        return True
    
    #  SortedInsert(node): Inserts the node into its proper position in a sorted list
    def sortedInsert(self, newNode):
        if self.isSorted() == False:
            self.sort()

        if self.head is None:
            self.insertHead(newNode)
        elif newNode.data < self.head.data:
            self.insertHead(newNode)
        else:
            current_node = self.head
            while current_node.next is not None and current_node.next.data < newNode.data: # We should insert the newNode in the spot after the node this loop stops at
                current_node = current_node.next
            
            # Insert the item at the found position
            newNode.prev = current_node
            newNode.next = current_node.next

            if current_node.next is not None:
                current_node.next.prev = newNode
            else:
                self.tail = newNode

            current_node.next = newNode

            self.size += 1
            
        
    
    def search(self, newNode):
        if self.size == 0:
            return None

        current = self.head
        for i in range(self.size):
            if current.data == newNode.data:
                return current
            else:
                current = current.next

        return None
    
    def deleteHead(self):
        if self.head is None:
            print("Error. There are no items in the list.")
        else:
            self.head = self.head.next # Set the head to the next item in the list (or none)

            if self.head is not None:
                self.head.prev = None # So it's not pointing to the old head
            else:
                self.tail = None # So that tail is not corrupt
            
            self.size -= 1
    
    def deleteTail(self):
        if self.tail is None:
            print("Error. There are no items in the list.")
        else:
            self.tail = self.tail.prev
            if self.tail is not None:
                self.tail.next = None
            else:
                self.head = None
            self.size -= 1
    
    def delete(self, delNode):
        # DOUBLE CHECK: JUST THE DATA HAS TO MATCH IN ORDER TO DELETE THIS
        # DOUBLE CHECK: IF THE DATA IS THE SAME FOR MANY NODES, DO WE DELETE ALL THE NODES OR JUST THE FIRST ONE?
        if delNode is None:
            return # If the deleted node is nothing
        if self.size == 0:
            return # If the list is empty
        
        else:
            current = self.head
            for i in range(0, self.size):
                if current == self.head and current.data == delNode.data:
                    self.deleteHead()
                elif current == self.tail and current.data == delNode.data:
                    self.deleteTail()
                elif current.data == delNode.data:
                    current.prev.next = current.next
                    current.next.prev = current.prev
                    self.size -= 1
                current = current.next
           
    
    def sort(self):
        if self.head is None or self.head.next is None:
            return

        sorted_tail = self.head  # tail of the sorted portion of the list
        current = self.head.next  # node to be inserted into the sorted portion

        while current is not None:
            if current.data < sorted_tail.data:
                # remove current from its current position
                current.prev.next = current.next
                if current.next is not None:
                    current.next.prev = current.prev

                # find the node to insert current before
                sorted_node = sorted_tail
                while sorted_node is not None and current.data < sorted_node.data:
                    sorted_node = sorted_node.prev

                # insert current before sorted_node
                if sorted_node is None:
                    # insert at the beginning
                    current.prev = None
                    current.next = self.head
                    self.head.prev = current
                    self.head = current
                else:
                    current.prev = sorted_node
                    current.next = sorted_node.next
                    sorted_node.next = current
                    if current.next is not None:
                        current.next.prev = current
            else:
                sorted_tail = current

            current = current.next
    
    def clear(self):
        self.head = None
        self.tail = None
        self.size = 0

    def print(self):
        print(f'The size of this doubly linked list is {self.size}.')
        if self.isSorted():
            print("This doubly linked list is sorted.")
        if not self.isSorted():
            print("This doubly linked list is not sorted.")
        print("This doubly linked list contains:")
        current = self.head
        for i in range(self.size):
            print(current.data)
            current = current.next


            
        
# Test Cases
# def main():  # Delete later
#     testDLL = doublyLL()
#     testNode1 = Node(1)
#     testNode2 = Node(2)
#     testNode3 = Node(3)
#     testNode4 = Node(4)
#     testNode5 = Node(5)   
#     testNode6 = Node(6)
#     testNode6_2 = Node(6)
#     testNode7 = Node(3)
#     testDLL.insertTail(testNode1)
#     testDLL.insertTail(testNode2)
#     testDLL.insertTail(testNode3) 
#     testDLL.insertTail(testNode4)
#     testDLL.insertTail(testNode5,)
#     testDLL.insertTail(testNode6)
#     testDLL.insertTail(testNode6_2)
#     testDLL.insertTail(testNode7)
#     testDLL.Print()


# if  __name__  == "__main__":
#     main()


import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Linear.SinglyLL import singlyLL
from datastructures.Nodes.SinglyNode import Node

testNode1 = Node(6)
testNode2 = Node(1)
testNode3 = Node(2)
testNode4 = Node(7)
testNode5 = Node(10)   
testNode6 = Node(3)
testNode6_2 = Node(3)
testNode7 = Node(4)

def setup():
    testSLL = singlyLL()
    testSLL.insertTail(testNode1)
    testSLL.insertTail(testNode2)
    testSLL.insertTail(testNode3) 
    testSLL.insertTail(testNode4)
    testSLL.insertTail(testNode5)
    testSLL.insertTail(testNode6)
    testSLL.insertTail(testNode6_2)
    testSLL.insertTail(testNode7)
    return testSLL

@pytest.mark.test_constructor
def test_constructor():
    testSLL = singlyLL()
    assert testSLL.head == None
    assert testSLL.tail == None
    assert testSLL.size == 0

    testSLL2 = singlyLL(testNode1)
    assert testSLL2.head == testNode1
    assert testSLL2.tail == testNode1
    assert testSLL2.size == 1

@pytest.mark.test_insert
def test_insert():
    testSLL = singlyLL()
    testSLL.insert(testNode1, 1)
    assert testSLL.head.data == 6
    testSLL.insertHead(testNode2)
    testSLL.insertTail(testNode3) 
    testSLL.insertTail(testNode4)
    assert testSLL.tail.data == 7
    testSLL.insert(testNode5, 3)
    assert testSLL.head.next.next.data == 10
    # Added these insert tests - test boundary cases
    testSLL.insert(Node(0), 1)
    assert testSLL.head.data == 0
    testSLL.insert(Node(0), 7)
    assert testSLL.tail.data == 0


    testSLL.insert(testNode6, 5)
    testSLL.insert(testNode7, 1) # test insert at head
    currentTestNode = testSLL.head 
    testList = []
    for i in range(0, testSLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4, 0, 1, 6, 10, 3, 2, 7, 0]
    testSLL.insertTail(testNode6_2)
    currentTestNode = testSLL.head 
    testList = []
    for i in range(0, testSLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next    
    assert testList == [4, 0, 1, 6, 10, 3, 2, 7, 0, 3]

@pytest.mark.test_delete
def test_delete():
    testNothing = singlyLL()
    testNothing.deleteHead()
    testNothing.deleteTail()
    assert testNothing.head == None
    assert testNothing.tail == None

    testNothing.insertTail(Node(1))
    testNothing.delete(Node(1))
    assert testNothing.head == None
    assert testNothing.tail == None
    assert testNothing.size == 0

    testNothing.insertTail(Node(1))
    testNothing.insertTail(Node(2))
    testNothing.delete(Node(2))

    assert testNothing.head.data == 1
    assert testNothing.tail.data == 1

    testNothing2 = singlyLL(Node(1))
    testNothing2.insertTail(Node(2))
    testNothing2.delete(Node(1))
    assert testNothing2.head.data == 2
    assert testNothing2.tail.data == 2

    testSLL = setup()
    testSLL.deleteHead()
    assert testSLL.head.data != 6
    testSLL.deleteTail()
    assert testSLL.tail.data != 4
    testSLL.delete(testSLL.tail)
    testSLL.delete(testNode6)
#*********************************
    # These 4 lines are new
    testSLL.insertHead(Node(11))
    testSLL.insert(Node(11),3)
    testSLL.insertTail(Node(11)) # insert 3 new duplicates
    testSLL.delete(Node(11))


    currentTestNode = testSLL.head
    testList = []
    for i in range(0, testSLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [1, 2, 7, 10]
    testSLL.delete(Node(10))
    testSLL.delete(Node(1))
    currentTestNode = testSLL.head
    testList = []
    for i in range(0, testSLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [2, 7]
    testSLL.clear()
    assert testSLL.size == 0 and testSLL.head == None and testSLL.tail == None

@pytest.mark.test_search
def test_search():
    testSLL = setup()
    testNode6Search = testSLL.search(testNode6)
    assert testNode6Search == testNode6
    nonListNode = Node(99)
    testNonListNode = testSLL.search(nonListNode)
    assert testNonListNode == None

@pytest.mark.test_sorting
def test_sorting():
    testSLL = setup()
    testSLL.sort()
    testList = []
    currentTestNode = testSLL.head
    for i in range(0, testSLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList == [1, 2, 3, 3, 4, 6, 7, 10]

    testIsSorted = testSLL.isSorted()

    assert testIsSorted == True

    newTestNode = Node(8)
    duplicateTestNode = Node(4)
    # Added sorted inserts for boundary cases (heads and tails) -- Added sortedInsert for Node(11) and Node(0)
    testSLL.sortedInsert(newTestNode)
    testSLL.sortedInsert(duplicateTestNode)
    testSLL.sortedInsert(Node(11))
    testSLL.sortedInsert(Node(0))
    currentTestNode = testSLL.head
    testList2 = []
    for i in range(0, testSLL.size):
        testList2.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList2 == [0, 1, 2, 3, 3, 4, 4, 6, 7, 8, 10, 11]

@pytest.mark.test_print
def test_print():
    testSLL = setup()
    captured_output = io.StringIO()
    sys.stdout = captured_output
    testSLL.print()
    actual_output = captured_output.getvalue()

    sys.stdout = sys.__stdout__

    expected_output = "The size of this singly linked list is 8.\nThis singly linked list is not sorted.\nThis singly linked list contains:\n6\n1\n2\n7\n10\n3\n3\n4\n"
    assert actual_output ==  expected_output


import pytest
import sys
import io
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from datastructures.Linear.DoublyLL import doublyLL
from datastructures.Nodes.DoublyNode import Node

testNode1 = Node(6)
testNode2 = Node(1)
testNode3 = Node(2)
testNode4 = Node(7)
testNode5 = Node(10)   
testNode6 = Node(3)
testNode6_2 = Node(3)
testNode7 = Node(4)

@pytest.mark.test_constructor
def test_constructor():
    testDLL = doublyLL()
    assert testDLL.head == None
    assert testDLL.tail == None
    assert testDLL.size == 0
    testDLL2 = doublyLL(Node(4))
    assert testDLL2.head.data == 4
    assert testDLL2.tail.data == 4
    assert testDLL2.size == 1


@pytest.mark.test_insert
def test_insert():
    testDLL = doublyLL()
    testDLL.insertHead(testNode1)
    assert testDLL.head.data == 6
    testDLL.insertHead(testNode2)
    testDLL.insertTail(testNode3) 
    testDLL.insertTail(testNode4)
    assert testDLL.tail.data == 7
    testDLL.insert(testNode5, 3)
    assert testDLL.head.next.next.data == 10

    # #NEW
    testDLL.insert(Node(0), 1)
    assert testDLL.head.data == 0
    testDLL.insert(Node(2), 7)
    assert testDLL.tail.data == 2
    
    testDLL.insert(testNode6, 5)
    testDLL.insert(testNode7,1) # test insert at head
    currentTestNode = testDLL.head 
    testList = []
    for i in range(0, testDLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [4, 0, 1, 6, 10, 3, 2, 7, 2]
    testDLL.insertTail(testNode6_2)
    testDLL.insert(Node(10), 11)
    currentTestNode = testDLL.head 
    testList = []
    for i in range(0, testDLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next    
    assert testList == [4, 0, 1, 6, 10, 3, 2, 7, 2, 3, 10]

    
def setup():
    testDLL = doublyLL()
    testDLL.insertTail(testNode1)
    testDLL.insertTail(testNode2)
    testDLL.insertTail(testNode3) 
    testDLL.insertTail(testNode4)
    testDLL.insertTail(testNode5)
    testDLL.insertTail(testNode6)
    testDLL.insertTail(testNode6_2)
    testDLL.insertTail(testNode7)
    return testDLL

@pytest.mark.test_isDoubly
def test_isDoubly():
    testDLL = setup()
    nextNode = testDLL.head.next
    prevNode = nextNode.prev
    assert prevNode.data == testDLL.head.data

@pytest.mark.test_delete
def test_delete():
    testDLL = setup()
    testDLL.deleteHead()
    assert testDLL.head.data != 6
    testDLL.deleteTail()
    assert testDLL.tail.data != 4
    testDLL.delete(testDLL.tail)
    testDLL.delete(testNode6)
    testDLL.insertHead(Node(11))
    testDLL.insert(Node(11), 3)
    testDLL.insertTail(Node(11))
    testDLL.insertTail(Node(11)) # insert 3 new duplicates
    testDLL.delete(Node(11))
    currentTestNode = testDLL.head
    testList = []
    for i in range(0, testDLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next
    assert testList == [1,2,7,10]
    testDLL.clear()
    assert testDLL.size == 0 and testDLL.head == None and testDLL.tail == None
        
@pytest.mark.test_search
def test_Search():
    testDLL = setup()
    testNode6Search = testDLL.search(testNode6)
    assert testNode6Search == testNode6
    nonListNode = Node(99)
    testNonListNode = testDLL.search(nonListNode)
    assert testNonListNode == None

@pytest.mark.test_sorting
def test_Sorting():
    testDLL = setup()
    testDLL.sort()
    testList = []
    currentTestNode = testDLL.head
    for i in range(0, testDLL.size):
        testList.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList == [1,2,3,3,4,6,7,10]

    testIsSorted = testDLL.isSorted()

    assert testIsSorted == True

    newTestNode = Node(8)
    duplicateTestNode = Node(4)
    # Current size of the list is 8
    testDLL.sortedInsert(newTestNode)
    testDLL.sortedInsert(duplicateTestNode)
    testDLL.sortedInsert(Node(11)) #new
    testDLL.sortedInsert(Node(0)) #new
    currentTestNode = testDLL.head
    testList2 = []
    for i in range(0, testDLL.size):
        testList2.append(currentTestNode.data)
        currentTestNode = currentTestNode.next

    assert testList2 == [0, 1, 2, 3, 3, 4, 4, 6, 7, 8, 10, 11]


@pytest.mark.test_print
def test_print():
    testDLL = setup()
    captured_output = io.StringIO()
    sys.stdout = captured_output
    testDLL.print()
    actual_output = captured_output.getvalue()

    sys.stdout = sys.__stdout__

    expected_output = "The size of this doubly linked list is 8.\nThis doubly linked list is not sorted.\nThis doubly linked list contains:\n6\n1\n2\n7\n10\n3\n3\n4\n"
    assert actual_output ==  expected_output


from setuptools import setup, find_packages

setup(name='redpjpj',
        version='1.0',
        description='ENSF 338 Final Project',
        author='Liam and Theodore',
        packages=find_packages(),
        install_requires=[]
        )
        


import sys
from pathlib import Path
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Nodes.SinglyNode import Node

class singlyLL():

    def __init__(self, newNode = None):
        if newNode is not None:
            self.head = newNode
            self.tail = newNode
            self.size = 1
        else:
            self.head = None
            self.tail = None
            self.size = 0
            
    def insertHead(self, newNode):
        if self.head is None:
            self.head = newNode
            self.tail = newNode
            self.size += 1
        else:
            newNode.next = self.head
            self.head = newNode
            self.size += 1
    
    def insertTail(self, newNode):
        if self.tail is None:
            self.head = newNode
            self.tail = newNode
            self.size += 1
        else:
            # newNode.tail = None
            self.tail.next = newNode
            self.tail = newNode
            self.size += 1

    def insert(self, newNode, position):
        if position < 1 or position > (self.size + 1):
            raise ValueError("Invalid position")
        
        if position == 1:
            self.insertHead(newNode)

        elif position == self.size + 1:
            self.insertTail(newNode)

        else:
            current = self.head
            for i in range(1, position - 1):
                current = current.next
            newNode.next = current.next
            current.next = newNode
            self.size += 1

    def sortedInsert(self, newNode):
        sortNeeded = self.isSorted()
        if sortNeeded == False:
            self.sort()

        if self.size == 0:
            self.insertHead(newNode)

        else:
            if newNode.data < self.head.data:
                self.insertHead(newNode)
            elif newNode.data > self.tail.data:
                self.insertTail(newNode)
            else:
                current = self.head.next
                while current.next != self.tail and current.next.data < newNode.data:
                    current = current.next
                newNode.next = current.next
                current.next = newNode
                self.size += 1
                    # if newNode.data > current.data :          
                    #     newNode.next = current.next
                    #     current.next = newNode
                    #     self.size += 1
                    #     break
                    # current = current.next
            
    def isSorted(self):
        if self.head is None or self.head.next is None:
            return True
        current = self.head
        for i in range(self.size-1):
            if current.next.data < current.data:
                return False
            current = current.next
        return True

    def sort(self):
        if self.head == None or self.head.next == None:
            return
        sorted_tail = self.head
        current = self.head.next        
        for i in range(self.size-1):
            if current.data < sorted_tail.data:
                sorted_tail.next = current.next # Cutting the current node out of the list 
               
                if current.data < self.head.data: # If the current node should be inserted at the beginning
                    current.next = self.head
                    self.head = current
                
                else:
                    sorted_node = self.head # Setting an additional pointer to the beginning of the list

                    while sorted_node != sorted_tail and current.data > sorted_node.next.data:
                        sorted_node = sorted_node.next # Iterates forwards and stops at a node where the current should be inserted after
                    current.next = sorted_node.next
                    sorted_node.next = current
            else:
                sorted_tail = current
            current = sorted_tail.next

        current = self.head
        for i in range(self.size-1): # better fix probably
            current = current.next
        self.tail = current

    def search(self, newNode):
        if self.size == 0:
            return None

        current = self.head
        for i in range(self.size):
            if current.data == newNode.data:
                return current
            else:
                current = current.next

        return None

    def deleteHead(self):
        if self.head is None:
            return
        elif self.head.next is None:
            self.head = None
            self.tail = None
            self.size -= 1
        else:
            self.head = self.head.next
            self.size -= 1

    def deleteTail(self):
        if self.tail is None:
            return
        current = self.head

        if current.next is None:
            self.tail = None
            self.head = None
            self.size -= 1
        while current.next != self.tail:
            current = current.next
        current.next = None
        self.tail = current
        self.size -= 1

    def delete(self, delNode):
        if self.size == 0:
            return
        elif self.size == 1:
            if self.head.data == delNode.data:
                self.head = None
                self.tail = None
                self.size -= 1
        else:
            current = self.head
            for i in range(1, self.size):
                if current == self.head and current.data == delNode.data:
                    self.head = current.next
                    self.size -= 1
                
                if current.next == None:
                    return
                
                elif current.next == self.tail:
                    if current.next.data == delNode.data:
                        self.deleteTail()
                    return
                
                elif current.next.data == delNode.data:
                    current.next = (current.next).next
                    current = current.next
                    self.size -= 1
                else:
                    current = current.next
    
    def clear(self):
        self.head = None
        self.tail = None
        self.size = 0

    def print(self):
        print(f'The size of this singly linked list is {self.size}.')
        if self.isSorted():
            print("This singly linked list is sorted.")
        if not self.isSorted():
            print("This singly linked list is not sorted.")
        print("This singly linked list contains:")
        current = self.head
        while current != None:
            print(current.data)
            current = current.next
        
# def main():  # Delete later
#     testNode1 = Node(0)
#     testNode2 = Node(1)
#     testNode3 = Node(6)
#     testNode4 = Node(10)
#     testNode5 = Node(2)
#     testNode6 = Node(6)

#     testLL = singlyLL()
#     testLL.insertTail(testNode1)
#     testLL.insertTail(testNode2)
#     testLL.insertTail(testNode3)
#     testLL.insertTail(testNode4)
#     testLL.insertTail(testNode6)
#     testLL.insertTail(testNode5)
#     testLL.delete(testNode6)
#     testLL.print()
# if  __name__  == "__main__":
#     main()


            



   
        



            

            







from pathlib import Path
import sys
parent_dir = Path(__file__).resolve().parent.parent
sys.path.append(str(parent_dir))
from Nodes.TNode import TNode

class BST:
    def __init__(self, arg = None):
        if arg == None:
            self.root = None
        elif isinstance(arg, int):
            self.root = TNode(arg)
        else: # If the argument is a TNode
            self.root = arg

    def setRoot(self, newRoot):
        nodesList = []
        def traverse(node):
            if node is not None:
                traverse(node.leftChild)
                nodesList.append(node)
                traverse(node.rightChild)
        traverse(self.root)

        if isinstance(newRoot, int):
            self.root = TNode(newRoot)
        elif isinstance(newRoot, TNode):
            self.root = newRoot
    
        for node in nodesList:
            self.insert(node)

    def getRoot(self):
        return self.root

    def insert(self, arg): 
        if isinstance(arg, int):
            node = TNode(arg)
        else: # If the argument is a TNode
            node = arg
        data = node.data
        
        current = self.root
        parent = None

        while current is not None:
            parent = current
            if data <= current.data:
                current = current.leftChild
            else:
                current = current.rightChild
        
        if self.root is None:
            self.root = node
        elif data <= parent.data:
            parent.leftChild = TNode(data, parent)
        else:
            parent.rightChild = TNode(data, parent)
    
    # NOTE: NEED TO WRITE UNIT TESTS FOR THE BOUNDARY CASE WHERE THERE ARE MANY NODES WITH THE SAME VALUE
    # Finds the node with val as data and delete it, if not found prints a statement that the value is not in the tree.
    def delete(self, value):
        # if self.root:
        while self.search(value):
            self.root = self._delete(value, self.root)
    
    def _delete(self, value, node): # HELPER METHOD
        if not node:
            return None
        
        if value < node.data:
            node.leftChild = self._delete(value, node.leftChild)
        elif value > node.data:
            node.rightChild = self._delete(value, node.rightChild)
        else:
            if not node.leftChild:
                return node.rightChild
            elif not node.rightChild:
                return node.leftChild
            else:
                min_node = self._find_min_node(node.rightChild)
                node.data = min_node.data
                node.rightChild = self._delete(min_node.data, node.rightChild)
        return node
    
    def _find_min_node(self, node): # HELPER METHOD
        while node.leftChild:
            node = node.leftChild
        return node


    def search(self, val):
        current = self.root
        while current is not None:
            if val == current.data:
                return current
            elif val < current.data:
                current = current.leftChild
            else:
                current = current.rightChild

    def printInOrder(self):
        def traverse(node):
            if node is not None:
                traverse(node.leftChild)
                print(node.data)
                traverse(node.rightChild)
        traverse(self.root)

    def printBF(self):
        queue = []
        queue.append(self.root)
        current_level_count = 1  # number of nodes in the current level
        next_level_count = 0  # number of nodes in the next level
        while len(queue) > 0:
            current = queue.pop(0) # Remove from the front of the queue
            print(current.data, end=' ')  # print without new line
            current_level_count -= 1
            if current.leftChild is not None:
                queue.append(current.leftChild)
                next_level_count += 1
            if current.rightChild is not None:
                queue.append(current.rightChild)
                next_level_count += 1
            if current_level_count == 0:
                # All nodes in the current level have been printed
                # Move to the next line and update counts
                print()
                current_level_count = next_level_count
                next_level_count = 0

    

def main():
    testNode1 = TNode(6)
    testNode2 = TNode(4)
    testNode3 = TNode(9)
    testNode4 = TNode(2)
    testNode5 = TNode(5)   
    testNode6 = TNode(7)
    testNode7 = TNode(3)
    testBST = BST(testNode1)
    testBST.insert(testNode2)
    testBST.insert(9)
    testBST.insert(2)
    testBST.insert(5)
    testBST.insert(testNode6)
    testBST.insert(testNode7)
    testBST.insert(2)
    testBST.insert(2)
    testBST.printBF()

if  __name__  == "__main__":
    main()    